# Narzędzia i Aplikacje Jakarta EE

## Platforma Jakarta EE

Michał Wójcik

---

### Aplikacje klasy enterprise - wymagania

**Wysoki poziom złożoności**:

* mnogość modułów,
* rozległa funkcjonalność:
  * tysiące wymagań biznesowych,
  * powiązania między modułami.

**Krytyczne dla działania przedsiębiorstwa:**

* wysokie wymagania w zakresie:
  * niezawodności,
  * wiarygodności,
  * bezpieczeństwa.

**Wymagania w zakresie wydajności:**

* dane wielkiej skali,
* wielu użytkowników/klientów.

---

### Aplikacje klasy enterprise - rozwój

**Rozwijane przez duży zespół:**

* setki deweloperów,
* konieczność wyodrębniania modułów i określania kontraktów pomiędzy nimi.

**Długoterminowy rozwój i utrzymanie:**

* perspektywa kilku-kilkunastu lat,
* inkrementalne dodawanie funkcjonalności,
* podejmowane decyzje mają długotrwałe skutki.

---

### Aplikacje klasy enterprise - integracja

**Wiele aplikacji klienckich:**

* przeglądarka internetowa,
* aplikacje mobilne,
* dedykowane aplikacje okienkowe,
* inne systemy (np. klientów biznesowych).

**Integracja z wieloma systemami:**

* zarówno wewnętrznymi, jak i zewnętrznymi,
* wiele różnorodnych interfejsów:
  * nierozwijanych lub przestarzałych,
  * o nieopłacalnym koszcie aktualizacji.

---

### Aplikacje klasy enterprise - technologia

**Możliwość testowania aplikacji na wielu poziomach:**

* testy jednostkowe,
* testy integracyjne,
* testy akceptacyjne,
* testy interfejsu użytkownika.

**Specyficzne wymagania technologiczne:**

* asynchroniczna wymiana komunikatów,
* komunikacja z gwarancją dostarczenia wiadomości,
* rozproszone transakcje,
* przetwarzanie wsadowe,
* zadania cykliczne.

---

### Aplikacje klasy enterprise - wytwarzanie

**Specyficzny model programistyczny:**

* promujący podział na komponenty,
* promujący luźne powiązania między składowymi,
* promujący podział na warstwy,
* pozwalający na definiowanie kontraktów pomiędzy komponentami,
* preferowane języki z silnym typowaniem.

---

### Jakarta EE

**Jakarta Enterprise Edition** - (Jakarta EE) zbiór specyfikacji rozszerzających Java SE służących
budowie aplikacji klasy *enterprise*.

Historia **wersji**:

* **Java 2 Platform, Enterprise Edition** (Sun Microsystems):
  * J2EE 1.2 (1999),
  * J2EE 1.3 (2001),
  * J2EE 1.4 (2003);
* **Java Enterprise Edition** (Sun Microsystems / Oracle Corporation):
  * Java EE 5 (2006),
  * Java EE 6 (2009),
  * Java EE 7 (2013),
  * Java EE 8 (2017);
* **Jakarta Enterprise Edition** (Eclipse Foundation):
  * Jakarta EE 8 (2019),
  * Jakarta EE 9 (2020),
  * Jakarta EE 10 (2022),
  * Jakarta EE 11 (w przygotowaniu).

---

### Poziomy Jakarta EE - układ

![Jakarta EE - warstwy](./assets/img/layers.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Poziomy Jakarta EE

**Poziomy** rozróżniane w ramach platformy Jakarta EE:

* **poziom klienta** - (client tier) uruchamiana na maszynie klienta (użytkownika):
  * samodzielne aplikacje okienkowe,
  * strony internetowe renderowane w przeglądarce;
* **poziom webowy** - (web tier) uruchamiana na serwerze aplikacji:
  * mechanizmy generowania dynamicznych stron internetowych, np.: JavaServer Faces (JSF) lub
    JavaServer Pages (JSP);
* **poziom biznesowy** - (business tier) uruchamiana na serwerze aplikacji:
  * implementacja logiki biznesowej,
* **poziom danych** - (Enterprise Information System) uruchomiona na serwerze EIS:
  * źródło danych.

---

### Poziomy Jakarta EE - układ elementów

![Jakarta EE - elementy warstw](./assets/img/layers-elements.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Poziomy Jakarta EE - elementy

**Elementy** rozróżniane poszczególnych poziomach platformy Jakarta EE:

* **poziom klienta**:
  * komponenty JavaBean dostarczający/reprezentujący dane,
  * strony internetowe renderowane przez przeglądarkę,
  * aplikacje klienckie,
  * aplikacje RIA (Rich Internet Application), np.: Applet (deprecated),
  * strony internetowe renderowane w przeglądarce;
* **poziom webowy**:
  * definicje widoków (stron internetowych), np.: JSF, JPS,
  * komponenty JavaBean dostarczające/reprezentujące dane,
  * komponenty przetwarzania żądań HTTP, np.: Servlet.
* **poziom biznesowy**:
  * obiekty reprezentujące tabele w bazie danych, np.: encje Java Persistence API (JPA),
  * komponenty logiki biznesowej, np.: Enterprise Java Beans (EJB) Session Beans,
  * komponenty sterowne komunikatami, np.: EJB Message-Driven Beans;
* **poziom danych**:
  * baza danych,
  * system ERP (Enterprise Resource Planing),
  * ...

---

### Kontenery Jakarta EE - układ

![Jakarta EE - kontenery](./assets/img/containers.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Kontenery Jakarta EE

**Kontener** - elementy platformy Jakarta EE sterujące wykonaniem komponentów znajdujących się na poszczególnych
poziomach.

W ramach platformy Jakarta EE można wyróżnić następujące **kontenery**:

* **kontener webowy** - uruchomiony po stronie serwera aplikacji, steruje cyklem życia servletów i stron JSF/JSP,
* **kontener EJB** - uruchomiony po stronie serwera aplikacji, steruje cyklem życia wszystkich komponentów EJB,
* **kontener kliencki** - uruchomiony po stronie klienta, steruje wykonaniem aplikacji klienckiej,
* **kontener apletów** - uruchomiony po stronie klienta w kontekście przeglądarki internetowej, steruje wykonaniem
  apletów (deprecated) osadzonych na stronach internetowych.

---

### Relacje pomiędzy kontenerami - układ

![Jakarta EE - relacje między kontenerami](./assets/img/containers-relation.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Relacje pomiędzy kontenerami

W typowej aplikacji Jakarta EE występują następujące **relacje** pomiędzy **kontenerami**:

* tylko komponenty EJB (kontener EJB) powinny mieć dostęp do źródła danych (np.: baza danych),
* komponenty JSF/JSP i servlety (kontener webowy) powinny wywoływać odpowiednie metody z komponentów
  EJB w celu zrealizowania operacji biznesowych,
* strony internetowe wyświetlane przez przeglądarkę są generowane przez komponenty JSF/JSP,
* dodatkowo żądania HTTP mogą być realizowane przez servlety,
* aplikacje klienckie (kontener kliencki) odwołują się bezpośrednio do komponentów EJB.

---

### Profile platformy Jakarta EE

**Profile Jakarta EE** określają jakie funkcje musi posiadać serwer aplikacji, aby uzyskać certyfikat zgodności z
platformą Jakarta EE.

**Full Profile** (Platform):

* wszystkie specyfikacje wchodzące w skład Jakarta EE.
* obsługa archiwów EAR (Enterprise Archive) oraz WAR.

**Web Profile** (od Java EE 6):

* profil przeznaczony dla aplikacji WWW (stron internetowych),
* nie obejmuje m.in.: EJB remoting, async, timer, JMS, Batch,
* tylko archiwa WAR (Web Archive).

**Core Profile** (od Jakarta EE 10):

* profil przeznaczony dla małych aplikacji opartych na usługach typu REST.

---

### Profile platformy Jakarta EE - web

**Web Profile (Jakarta EE 10)**:

* Jakarta Annotations 2.1
* Jakarta Authentication 3.0
* Jakarta Bean Validation 3.0
* Jakarta Concurrency 3.0
* Jakarta Contexts and Dependency Injection 4.0
* Jakarta Debugging Support for Other Languages 2.0
* Jakarta Dependency Injection 2.0
* Jakarta Enterprise Beans 4.0 Lite
* Jakarta Expression Language 5.0
* Jakarta Interceptors 2.1
* Jakarta JSON Binding 3.0
* Jakarta JSON Processing 2.1
* Jakarta Persistence 3.1
* Jakarta RESTful Web Services 3.1
* Jakarta Security 3.0
* Jakarta Server Faces 4.0
* Jakarta Server Pages 3.1
* Jakarta Servlet 6.0
* Jakarta Standard Tag Library 3.0
* Jakarta Transactions 2.0
* Jakarta WebSocket 2.1

---

### Profile platformy Jakarta EE - full

**Full Profile (Jakarta EE 10)**:

* Wymagane:
  * Jakarta Activation 2.1
  * Jakarta Authorization 2.1
  * Jakarta Batch 2.1
  * Jakarta Common Annotations 2.1
  * Jakarta Connectors 2.1
  * Jakarta Enterprise Beans 4.0 (except for Jakarta Enterprise Beans entity beans and associated Jakarta Enterprise Beans
    QL, and embedded container, which have been made removed)
  * Jakarta Mail 2.1
  * Jakarta Managed Beans 2.0
  * Jakarta Messaging 3.1
* Opcjonalne:
  * Jakarta Enterprise Beans 3.2 and earlier entity beans and associated Jakarta Enterprise Beans QL
  * Jakarta Enterprise Beans 2.x API group
  * Jakarta Enterprise Web Services 2.0
  * Jakarta SOAP with Attachments 3.0
  * Jakarta XML Web Services 4.0
  * Jakarta XML Binding 4.0

---

### Profile platformy Jakarta EE - core

**Core Profile (Jakarta EE 10)**:

* Jakarta Annotations 2.1
* Jakarta Contexts and Dependency Injection (CDI) 4.0 (CDI Lite section)
* Jakarta Dependency Injection 2.0
* Jakarta Interceptors 2.1
* Jakarta JSON Processing 2.1
* Jakarta JSON Binding 3.0
* Jakarta RESTful Web Services 3.1

---

### Kontenery webowe

**Web/Servlet container** - komponent serwera webowego zajmującego się interakcją z servletami i potencjalnie stronami
JSP.

**Apache Tomcat 10.1.x**:

* Jakarta Servlet 6.0,
* Jakarta Server Pages 3.1,
* ...

**Eclipse Jetty 12.0.x**:

* Jakarta Servlet 6.0,
* Jakarta Server Pages 3.1,
* ...

---

### Serwery aplikacji - Java EE 8

**Serwery aplikacji** oficjalnie kompatybilne z **Java EE 8**:

* GlassFish Server Open Source Edition 5.0,
* Eclipse GlassFish 5.1,
* Wildfly 14.x,
* Red Hat JBoss Enterprise Application Platform 7.2,
* IBM WebSphere Application Server Liberty 18.0.0.2.

**Serwery aplikacji** oficjalnie kompatybilne z **Jakarta EE 8**:

* Eclipse Foundation, Eclipse Glassfish 5.1.0,
* Payara Services Limited, Payara Server Community 5.2021.4,
* Payara Services Limited, Payara Server Enterprise 5.53.0,
* Red Hat, Wildfly 18.0.0,
* IBM, Open Liberty 20.0.0.3.

---

### Serwery aplikacji - Jakarta EE 9 i 10

**Serwery aplikacji** oficjalnie kompatybilne z **Jakarta EE 9.1**:

* Eclipse Foundation, Eclipse Glassfish 6.1,
* ManageCat, ManageFish Server 6.1.0,
* Payara Services Limited, Payara Server Community 6.2021.1 Alpha 1,
* Red Hat, Wildfly 25.0.0.Final,
* IBM, Open Liberty 21.0.0.12.

**Serwery aplikacji** oficjalnie kompatybilne z **Jakarta EE 10**:

* Eclipse Foundation, Eclipse Glassfish 7,
* Payara Services Limited, Payara Server Community 6.2023.7,
* Payara Services Limited, Payara Server Enterprise 6.4.0,
* Red Hat, Wildfly 27.0.0.Alpha5,
* IBM, Open Liberty 23.0.0.3.

---

### Struktura aplikacji - archiwum EAR

Archiwum **EAR**.

![Archiwum EAR](./assets/img/ear.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Struktura aplikacji - zawartość EAR

Archiwum **EAR** - archiwum zawierające aplikację typu enterprise złożoną z kilku modułów:

* pliki z rozszerzeniem `.ear`,
* archiwa typu ZIP.

**Zawartość** archiwum **EAR**

* `META-INF`:
  * `application.xml` - standardowy ogólny deskryptor z konfiguracją aplikacji,
  * `<server_name>-application.xml` - deskryptor z konfiguracją specyficzną dla danego serwera aplikacji;
* web module - dowolna liczba modułów webowych (archiwa WAR),
* application client module - dowolna liczba aplikacji klienckich (archiwa JAR),
* resource adapter module - dowolna liczba modułów realizujących połączeni z różnymi źródłami danych (archiwa JAR),
* EJB module - dowolna liczba modułów zawierających komponenty EJB (archiwa EJB-JAR).

---

### Struktura aplikacji - archiwum EJB-JAR

Archiwum **EJB-JAR**.

![Archiwum EJB-JAR](./assets/img/jar.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Struktura aplikacji - zawartość EJB-JAR

Archiwum **EJB-JAR** - archiwum zawierające samodzielne komponenty EJB:

* pliki z rozszerzeniem `.jar`,
* archiwa typu ZIP,
* może być wdrażane samodzielnie na serwer aplikacji,
* może być wykorzystywany przez inne aplikacje uruchomione na serwerze.

**Zawartość** archiwum **EJB-JAR**:

* `META-INF`:
  * `ejb-jar.xml` - standardowy ogólny deskryptor z konfiguracją aplikacji,
  * `<server_name>-ejb-jar.xml` - deskryptor z konfiguracją specyficzną dla danego serwera aplikacji;
* pliki `.class` - wszystkie klasy modułu.

---

### Struktura aplikacji - archiwum WAR

Archiwum **WAR**.

![Archiwum WAR](./assets/img/war.svg)

<https://eclipse-ee4j.github.io/jakartaee-tutorial/>

---

### Struktura aplikacji - zwartość WAR

Archiwum **WAR** - archiwum zawierające samodzielną aplikację webową:

* pliki z rozszerzeniem `.war`,
* archiwa typu ZIP,
* może być wdrażane samodzielnie na serwer aplikacji,
* może korzystać z innych aplikacji uruchomionych na serwerze.

**Zawartość** archiwum **WAR**:

* `WEB-INF`:
  * `web.xml` - standardowy ogólny deskryptor z konfiguracją aplikacji,
  * `<server_name>-web.xml` - deskryptor z konfiguracją specyficzną dla danego serwera aplikacji;
  * `classes` - zawiera wszystkie klasy modułu,
  * `lib` - biblioteki wymagane przez moduł (archiwa JAR);
* strony internetowe.

---

### Pierwsza aplikacja - konfiguracja

Ogólna struktura pliku `pom.xml`:

```xml
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0"
         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 
                             http://maven.apache.org/xsd/maven-4.0.0.xsd">

    <modelVersion>4.0.0</modelVersion>

    <groupId>pl.edu.pg.eti.kask</groupId>
    <artifactId>simple-rpg</artifactId>
    <version>1.0-SNAPSHOT</version>
    <packaging>war</packaging>

    <name>Simple RPG</name>
    <description>Simple RPG game</description>

    <dependencies>
        <!-- ... -->
    </dependencies>

    <build>
        <!-- ... -->
    </build>

</project>
```

---

### Pierwsza aplikacja - zależności

Zależność dla Jakarta EE musi być ustawiona jako `provided` ponieważ jest dostarczana po stronie serwera.

```xml
<dependency>
    <groupId>jakarta.platform</groupId>
    <artifactId>jakarta.jakartaee-api</artifactId>
    <version>10.0.0</version>
    <scope>provided</scope>
</dependency>
```

---

### Pierwsza aplikacja - archiwum

Wymagany jest plugin generujący archiwum `war`.

```xml
<plugins>
    <plugins>
        <plugin>
            <groupId>org.apache.maven.plugins</groupId>
            <artifactId>maven-war-plugin</artifactId>
            <version>3.3.2</version>
        </plugin>
    </plugins>
</plugins>
```

---

### Pomocne źródła - Jakarta EE 8 - 10

Jakarta EE 10:

* Eclipse Foundation, *Jakarta EE Specifications*,
  <https://jakarta.ee/specifications/>.
* Eclipse Foundation, *Your First Cup: An Introduction to the Jakarta EE Platform Examples*,
  <https://github.com/eclipse-ee4j/jakartaee-firstcup-examples>.
* Eclipse Foundation, *Jakarta EE 10 Examples*,
  <https://github.com/eclipse-ee4j/jakartaee-examples>.

Jakarta EE 9:

* Eclipse Foundation, *Jakarta EE 9 Tutorial*,
  <https://eclipse-ee4j.github.io/jakartaee-tutorial/>.

Jakarta EE 8:

* Eclipse Foundation, *Your First Cup: An Introduction to the Jakarta EE Platform*,
  <https://eclipse-ee4j.github.io/jakartaee-firstcup/toc.html>.

---

### Pomocne źródła - Java EE 7 - 8

Java EE 8:

* Oracle, *Java EE 8 Tutorial*, <https://javaee.github.io/tutorial/toc.html>.
* Oracle, *Your First Cup: An Introduction to the Java EE Platform*,
  <https://javaee.github.io/firstcup/toc.html>.
* Oracle, *Java EE 8 Tutorial Examples*, <https://github.com/javaee/tutorial-examples>.
* Oracle, *Your First Cup: An Introduction to the Java EE Platform Examples*,
  <https://github.com/javaee/firstcup-examples>.

Java EE 7:

* WildFly Team, *Java EE 7 Tutorial*, <https://docs.wildfly.org/17/JavaEE_Tutorial.html>.
* Oracle, *Java EE 7 Tutorial*, <https://docs.oracle.com/javaee/7/tutorial/>.
* Oracle, *Your First Cup: An Introduction to the Java EE Platform*,
  <https://docs.oracle.com/javaee/7/firstcup/>.

---

### Pomocne źródła - Java EE 5 - 6

Java EE 6:

* Oracle, *Java EE 6 Tutorial*, <https://docs.oracle.com/javaee/6/tutorial/doc/>.
* Oracle, *Your First Cup: An Introduction to the Java EE Platform*,
  <https://docs.oracle.com/javaee/6/firstcup/doc/>.

Java EE 5:

* Oracle, *Java EE 5 Tutorial*, <https://docs.oracle.com/javaee/5/tutorial/doc/>.
* Oracle, *Your First Cup: An Introduction to the Java EE Platform*,
  <https://docs.oracle.com/javaee/5/firstcup/doc/>.

---

### Pomocne źródła - pozostałe

Kursy:

* Baeldung , *Tutorials*, <https://www.baeldung.com/category/java-ee/>.
