# Narzędzia i Aplikacje Jakarta EE

## Jakarta Faces - Zagadnienia Zaawansowane

Michał Wójcik

---

### Protokół WebSocket

**WebSocket** - protokół komunikacyjny w ramach połączenia TCP stworzony z myślą o komunikacji pomiędzy przeglądarką (
lub innym klientem) i serwerem webowym.

**WebSocket** w Java EE:

* od Java EE 7:
  * Java API for WebSocket (JSR 356),
  * bazuje na implementacji z OmniFaces,
  * biblioteka OmniFaces ułatwia wykorzystanie mechanizmów serwera.

Uwaga: Mojarra (implementacja JSF wykorzystana w serwerze WildFly) nie obsługuje poprawnie WebSocket po ponownym
wdrożeniu aplikacji: <https://github.com/eclipse-ee4j/mojarra/issues/4572>.

---

### WebSocket - konfiguracja

Konfiguracja w `web.xml`:

```xml
<context-param>
    <param-name>jakarta.faces.ENABLE_WEBSOCKET_ENDPOINT</param-name>
    <param-value>true</param-value>
</context-param>
```

---

### WebSocket by OmniFaces - konfiguracja

Zależność:

```xml
<dependency>
    <groupId>org.omnifaces</groupId>
    <artifactId>omnifaces</artifactId>
    <version>2.6.5</version>
</dependency>
```

Konfiguracja w `web.xml`:

```xml
<context-param>
    <param-name>org.omnifaces.SOCKET_ENDPOINT_ENABLED</param-name>
    <param-value>true</param-value>
</context-param>
```

---

### Emitowanie zdarzeń

```jshelllanguage
@Inject
@Push
PushContext bookUpdates;

public void notify() {
    bookUpdates.send("update");
}
```

Uwaga: Nazwa kanału WebSocket domyślnie określana na podstawie nazwy pola.

---

### Nasłuchiwanie zdarzeń - JavaScript

Po prostu JavaScript:

```javascript
let socket = new WebSocket("ws://localhost:8080/library/javax.faces.push/bookUpdates");
// let socket = new WebSocket("wss://localhost:8443/library/javax.faces.push/bookUpdates");

socket.onopen = function (e) {
};

socket.onmessage = function (event) {
    console.log(event.data);
};

socket.onclose = function (event) {
};

socket.onerror = function (error) {
};
```

---

### Nasłuchiwanie zdarzeń - JSF

JSF:

```xml
<f:websocket channel="bookUpdates" onmessage="receive"/>
```

OmniFaces:

```xml
<o:socket channel="bookUpdates" onmessage="receive"/>
```

Funkcja JavaScript, wykorzystana w komponentach JSF:

```javascript
function receive(msg, channel, event) {
    var data = JSON.parse(event.data);
    var book = data.data;
}
```

---

### WebSocket

**WebSocket**:

* mechanizm umożliwia dwustronną komunikację
  * z serwera do przeglądarki,
  * z przeglądarki do serwera;
* nie zaleca się używania do wysyłania informacji z przeglądarki do serwera:
  * lepiej wysłać dane w żądaniu HTTP za pomocą AJAX,
  * takie żądanie podlega takiej samej obsłudze jak wszystkie inne żądania HTTP obsługiwane przez serwer,
  * zawiera cookies oraz parametr viewState,
  * podlega mechanizmom kontroli dostępu, filtrom itd.

---

### AJAX

**AJAX** - Asynchronous JavaScript and XML, technika wytwarzania aplikacji webowych bazująca na wykonywaniu w tle
asynchronicznych żądań do serwera i modyfikacji wyświetlanego dokumentu bez przeładowywania całości:

* umożliwia budowę interaktywnych interfejsów dla aplikacji,
* JSF posiada wbudowaną obsługę dla wszystkich standardowych komponentów,
* żądania opierają się na mechanizmie partial processing (częściowe przetwarzanie widoku)
  * należy określić komponenty, których wartości zostaną przesłane w żądaniu,
  * oraz komponenty, które mają zostać zaktualizowane w przeglądarce po otrzymaniu odpowiedzi.

---

### Partial Processing

```xml
<h:commandButton value="Delete" actionListener="#{listBooks.remove(book)}">
    <f:ajax execute="@form" render=":books"/>
</h:commandButton>
```

Atrybuty:

* `execute` – co ma zostać przesłane na serwer:
  * należy wskazań komponenty według ich id (client ID),
  * można wymienić wiele komponentów rozdzielając je spacjami,
  * specjalne wartości: `@this` (domyślne), `@form`, `@all`, `@none`;
* `render` – co ma zostać odświeżone na stronie:
  * wartości takie same jak dla execute.

---

### Client ID

**Client ID** - wartość id komponenty html wyrenderowanego przez FacesServlet:

* obecność naming containers wpływa na wyliczoną wartość.

```xml
<body>
    <h:inputText id="input"
                 value="#{viewBean.input}"/>
</body>
```

```html
<body>
    <input id="input"
           type="text" name="input">
</body>
```

```xml
<body>
    <h:form id="form">
        <h:inputText id="input"
                     value="#{viewBean.input}"/>
    </h:form>
</body>
```

```html
<body>
    <form id="form" name="form" ...>
        <input id="form:input"
               type="text" name="form:input">
    </form>
</body>
```

Naming container - komponent JSF implementujący interfejs `NamingContainer`, np.:

* `HtmlDataTable`,
* `HtmlForm`.

---

### i18n w JSF

**i18n** - **i**internationalizatio**n** w JSF pozwala na:

* zdefiniowanie pakietów zasobów (resource bundle) używanych do wyświetlania niestandardowych wiadomości,
* zdefiniowanie pakietów wiadomości (message bundle) używanych do nadpisania standardowych wiadomości,
* zdefiniowania zasobów (resources) wyświetlanych w różnych językach.

---

### Konfiguracja i18n w JSF

Konfiguracja wielu języków wymaga zdefiniowania w deskryptorze `faces-config.xml`:

* dostępnych ustawień językowych (language locale):
  * w postaci `<language code>[_<country code>[_<variant code>]]`,
  * np.: `en`, `en_US`, `en_UK`;
* domyślnego ustawienia językowego.

```xml
<?xml version='1.0' encoding='UTF-8'?>
<faces-config version="2.3">
  <application>
    <locale-config>
      <default-locale>en</default-locale>
      <supported-locale>en</supported-locale>
      <supported-locale>pl</supported-locale>
    </locale-config>
  </application>
</faces-config>
```

---

### Resource bundle

**Resource bundle** - pozwalają na przygotowanie zasobów dla różnych ustawień językowych:

* obiekty klasy `java.util.ResourceBundle`,
* mogą to być dowolne obiekty,
* najczęściej są to tłumaczenia wiadomości wykorzystywanych w interfejsie użytkownika:
  * realizowane jako pliki `.properties`.

---

### Resource bundle - konfiguracja

```xml
<?xml version='1.0' encoding='UTF-8'?>
<faces-config version="2.3">
  <application>
    <resource-bundle>
      <base-name>
        pl.edu.pg.eti.kask.javaee.example.library.i18n.messages
      </base-name>
      <var>msg</var>
    </resource-bundle>
  </application>
</faces-config>
```

Wartość `messages` wskazuje na pliki `.properties` które mogą być nazwane np.:

* `messages_pl.properties`,
* `messages_en_US.properties`,
* `messages_en_UK.properties`.

---

### Resource bundle - przykład

Tłumaczenia dla resource bundle są zwykłymi plikami properties:

`messages_en.properties`

```tex
books=books
book=book
```

`messages_pl.properties`

```tex
books=książki
book=książka
```

---

### Resource bundle - lokalne

Definiowanie zasobów lokalnych dla widoku:

```xml
<ui:define name="content">
    <f:loadBundle var="msg" basename="pl.edu.pg.eti.kask.javaee.example.i18n.msg"/>
</ui:define>
```

---

### Message bundle

**Message bundle** - specjalny rodzaj resource bundle pozwalający na nadpisanie domyślnych
wiadomości JSF.

```xml
<?xml version='1.0' encoding='UTF-8'?>
<faces-config version="2.3">
  <application>
    <message-bundle>pl.edu.pg.eti.kask.javaee.example.i18n.messages</message-bundle>
  </application>
</faces-config>
```

```tex
jakarta.faces.converter.DateTimeConverter.DATE = Błędny format daty
```

---

### Zasoby JSF - struktura

Statyczne zasoby JSF powinny znajdować się w katalogu `resources` i być podzielone na *biblioteki*,
np.:

```tex
|-resources
|--js
|---nav.js
|--img
|---header.png
|--css
|---main.css
```

lub:

```tex
|-resources
|--default
|---img
|----header.png
|---css
|----main.css
|--fancy
|---img
|----header.png
|---css
|----main.css
```

---

### Zasoby JSF - użycie

```xml
<h:head>
    <h:outputStylesheet library="css" name="main.css"/>
</h:head>
```

```xml
<h:head>
    <h:outputScript library="js" name="nav.js" target="body"/>
</h:head>
```

```xml
<h:head>
    <h:outputScript library="js" name="nav.js" target="head"/>
</h:head>
```

```xml
<h:body>
    <h:graphicImage library="img" name="header.png"/>
</h:body>
```

---

### Zasoby JSF - biblioteki

```xml
<h:head>
    <h:outputStylesheet library="default" name="css/main.css"/>
</h:head>
```

```xml
<h:body>
    <h:graphicImage library="default" name="img/header.png"/>
</h:body>
```

Dynamicznie:

```xml
<h:head>
    <h:outputStylesheet library="#{conf.style}" name="css/main.css"/>
</h:head>
```

```xml
<h:body>
    <h:graphicImage library="#{conf.style}" name="img/header.png"/>
</h:body>
```

---

### Tłumaczenie zasobów JSF

Biblioteki zasobów JSF można umieścić w katalogach wskazujących na ustawienia językowe:

```tex
|-resources
|--pl
|---img
|----header.png
|--en
|---img
|----header.png
```

Automatyczne dodawanie ustawień językowych do biblioteki wymaga wpisu w message bundle:

```properties
javax.faces.resource.localePrefix=pl
```

Użycie pozostaje bez zmian:

```xml
<h:body>
    <h:graphicImage library="img" name="header.png"/>
</h:body>
```

---

### Pomocne źródła

Specyfikacja:

* <https://jakarta.ee/specifications/faces/4.0/jakarta-faces-4.0>
* <https://jakarta.ee/specifications/websocket/2.1/jakarta-websocket-spec-2.1>
* <https://jakarta.ee/specifications/expression-language/5.0/jakarta-expression-language-spec-5.0>
