# Narzędzia i Aplikacje Jakarta EE

## Jakarta Server Pages

Michał Wójcik

---

### JavaServer Pages

**Jakarta Server Pages** - (JavaServer Pages) (JSP):

* strona JSP składa się ze statycznych (tekstowych) elementów i konstrukcji JSP, które tworzą część
  dynamiczną strony,
* JSP umożliwia stosowanie dwóch typów składni:
  * standardowa,
  * xml;
* zalecane rozszerzenia to `.jsp`, `.jspx` i `.jspf` (dla fragmentów stron),

---

### Strona JSP

```xml
<jsp:root xmlns:jsp="http://xmlns.jcp.org/JSP/Page"
          xmlns="http://www.w3.org/1999/xhtml"
          xmlns:c="http://xmlns.jcp.org/jsp/jstl/core"
          version="2.0">
    <jsp:directive.page contentType="text/html"/>
    <jsp:text>&lt;!DOCTYPE html&gt;</jsp:text>

    <html lang="en">
        <head>
            <meta charset="UTF-8"/>
            <title>Library</title>
        </head>
        <body>
            <c:forEach var="i" items="1,2,3">
                <p>${i}</p>
            </c:forEach>
        </body>
    </html>
</jsp:root>
```

---

### Cykl życia strony JSP

**Strony JSP**:

* są przetwarzane jako servlety, więc dziedziczą po nich wiele cech,
* kiedy następuje odwołanie do strony JSP, jest ona tłumaczona na kod servletu (kod źródłowy Javy), a następuje jego
  kompilacja.

**Cykl życia strony JSP**:

1. jeśli nie istnieje instancja servletu danej strony JSP, kontener:
    1. ładuje klasę servletu JSP,
    2. tworzy jej instancję,
    3. wywołuje metodę `jspInit()`;
2. wywoływana jest metoda `_jspService()` (odpowiada zawartości strony JSP),
3. gdy potrzebne jest usunięcie servletu JSP z pamięci wywoływana jest metoda `jspDestroy()`.

---

### JSP - elementy sterujące

Do wygenerowania dokumentu HTML często potrzebne są instrukcje pętli, instrukcje warunkowe itd.

JSP udostępnia dwa podstawowe rozwiązania w tym zakresie:

* scriptlety:
  * niezalecane,
  * umożliwiają osadzenie dowolnego kodu Javy pomiędzy znacznikami HTML w pliku JSP (taki PHP dla Javy),
  * niezalecane, ponieważ mogą prowadzić do pomieszania logiki obsługi żądania z widokiem;
* JSTL Core:
  * znaczniki sterujące o składni XML do zastosowania w plikach JSP.

Istnieje możliwość definiowanie własnych znaczników i funkcji.

---

### Java Standard Tag Library

**Java Standard Tag Library** (JSTL) - standardowa biblioteka znaczników dla JSP:

* Core (`core`):
  * `remove`,
  * `set`,
  * `choose`,
  * `forEach`,
  * `forTokens`,
  * `if`,
  * `import`,
  * `redirect`,
  * `url`,
  * `catch`,
  * `out`.
* i18n (`fmt`):
  * `setLocale`,
  * `requestEncoding`,
  * `bundle`,
  * `message`,
  * `setBundle`,
  * `formatNumber`,
  * `formatDate`,
  * `parseDate`,
  * `parseNumber`,
  * `setTimeZone`,
  * `timeZone`.
* Database (`sql`):
  * `setDataSource`,
  * `query`,
  * `transaction`,
  * `update`.
* XML (`xml`)
  * `out`,
  * `parse`,
  * `set`,
  * `choose`,
  * `forEach`,
  * `if`,
  * `transform`.

---

### JSTL - przykład

```xml
<jsp:root xmlns:jsp="http://java.sun.com/JSP/Page"
          xmlns="http://www.w3.org/1999/xhtml"
          xmlns:c="http://xmlns.jcp.org/jsp/jstl/core"
          version="2.0">
    <c:forEach var="i" items="1,2,3">
        <c:if test="${i lt 2}">
            <p>${i}</p>
        </c:if>
    </c:forEach>
</jsp:root>
```

---

### JSTL - funkcje

Funkcje wywoływane bezpośrednio ze strony JSP (`functions`):

* `length`,
* `toUpperCase`,
* `toLowerCase`,
* `substring`,
* `substringAfter`,
* `substringBefore`,
* `trim`,
* `replace`,
* `indexOf`,
* `startsWith`,
* `endsWith`,
* `contains`,
* `containsIgnoreCase`,
* `split`,
* `join`,
* `escapeXml`.

---

### JSTL - funkcje - przykład

```xml
<jsp:root xmlns:jsp="http://java.sun.com/JSP/Page"
          xmlns="http://www.w3.org/1999/xhtml"
          xmlns:c="http://java.sun.com/jsp/jstl/core"
          xmlns:f="http://java.sun.com/jsp/jstl/functions"
          version="2.0">
    <c:forEach var="i" items="a,aaa,aaa">
        <p>
            ${f:replace(i, 'a', 'b')}
        </p>
    </c:forEach>
</jsp:root>
```

---

### Współdzielenie obiektów

Kontener udostępnia mechanizm **współdzielenia** obiektów na różnych **poziomach**:

* aplikacji webowej,
* sesji,
* żądania,
* strony.

```xml
<jsp:useBean id="service" class="pl.edu.pg.eti.kask.javaee.example.Service"
             scope="application"/>
```

```xml
<jsp:useBean id="viewBean" class="pl.edu.pg.eti.kask.javaee.example.ViewBean"
             scope="page"/>
```

```xml
<jsp:setProperty name="viewBean" property="service" value="${service}"/>
```

---

### Logika obsługi żądań

Implementacja **logiki obsługi żądań**:

* umieszczanie logiki obsługi żądania w pliku JSP nie jest zalecane:
  * scriptlety, JSTL SQL itd;
* możliwe jest umieszczenie logiki w beanach dostępnych za pomocą `<jsp:useBean>`:
  * obsłuży jedynie wyświetlenie danych;
* JSP jako View w MVC (Model-View-Controller):
  * strony JSP realizują widok i wyświetlają dane dostępne w modelu (dostęp przez EL),
  * servlety pełnią funkcję kontrolerów, obsługują żądanie, umieszczają beany modelu w odpowiednim zasięgu i
    przekierowują na stronę JSP,
  * kontrolerem może być również bean JAX-RS (Jakarta MVC) lub `@Controller` ze Spring MVC.

---

### JSP dzisiaj

W nowych projektach JSP ustępuje miejsca nowocześniejszym rozwiązaniom

* Thymeleaf,
* Facelets,
* FreeMarker,
* Mustache
* ...

JSP jest ciągle obecne w starszych projektach

* które będą utrzymywane jeszcze przez wiele lat,
* wykorzystywane wyłącznie jako język budowania szablonów widoków,
* bez scriptletów, bez JSTL database i innych przestarzałych rozwiązań.

---

### Pomocne źródła

Specyfikacja:

* <https://jakarta.ee/specifications/pages/3.1/jakarta-server-pages-spec-3.1>

Uzupełnienie informacji:

* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#jakarta-server-pages-technology>
* <https://docs.oracle.com/javaee/5/tutorial/doc/bnagx.html>
* <https://www.baeldung.com/jsp>
* <https://www.baeldung.com/spring-boot-jsp>
* <https://www.baeldung.com/jsf-servlet-jsp>
