# Narzędzia i Aplikacje Jakarta EE

## TypeScript

Michał Wójcik

---

### Problemy JavaScript

Problemy z JavaScript - **dynamiczne typowanie**:

* brak weryfikacji typów na etapie wytwarzania kodu - dopiero na etapie uruchomienia w przeglądarce,
* ograniczone wsparcie narzędziowe (IDE nie jest w stanie prawidłowo podpowiadać dostępnych metod obiektów),
  * kontrakty funkcji i obiektów opisane w dokumentacji (JSDoc), a nie w kodzie źródłowym,
  * dokumentacja jest często nieaktualna, nikt nie lubi jej pisać;
* dynamiczne typowanie jest wygodne w małych projektach,
* w dużych, wieloosobowych projektach statyczne typowanie często ułatwia pracę:
  * kontrola typów na etapie kompilacji/transpilacji, lepsze wsparcie narzędziowe.

---

### Problemy JavaScript cd

**Problemy** z JavaScript:

* brak wsparcia dla modułów (pakietów, zasięgów) na poziomie języka,
* IIFE - immediately invoked function expression - wprowadzanie zasięgu leksykalnego przy użyciu zasięgu funkcji,
* brak granularnej kontroli nad widocznością pól obiektów (np. private, package, protected, public),
* w efekcie: JavaScript słabo skaluje się w dużych projektach.

---

### ECMAScript

**ECMAScript** - ustandaryzowany obiektowy skryptowy język programowania, którego implementacją jest JavaScript.

| wersja       | nazwa                          | rok  | uwagi                          |
|--------------|--------------------------------|------|--------------------------------|
| ES1          | ECMAScript 1                   | 1997 |                                |
| ES2          | ECMAScript 2                   | 1998 | zmiany edytorskie              |
| ES3          | ECMAScript 3                   | 1999 |                                |
| ES4          | ECMAScript 4                   |      | nie doczekał się publikacji    |
| ES5          | ECMAScript 5                   | 2009 | tu został Internet Explorer 11 |
| ES5          | ECMAScript 5.1                 | 2011 | zmiany edytorskie              |
| ES6 / ES2015 | ECMAScript 6 / ECMAScript 2015 | 2015 |                                |
| ES2016       | ECMAScript 2016                | 2016 |                                |
| ES2017       | ECMAScript 2017                | 2017 |                                |
| ES2018       | ECMAScript 2018                | 2018 |                                |
| ES2019       | ECMAScript 2019                | 2019 |                                |
| ES2020       | ECMAScript 2020                | 2020 |                                |
| ES2021       | ECMAScript 2021                | 2021 |                                |
| ES2022       | ECMAScript 2022                | 2022 |                                |
| ES2023       | ECMAScript 2023                | 2023 |                                |

---

### Język TypeScript

**TypeScript**:

* statycznie typowany język transpilowany do języka JavaScript
  * deweloper pracuje w języku TypeScript, przeglądarka otrzymuje zrozumiały dla siebie kod JavaScript;
* oferuje kompatybilność z najnowszymi wersjami języka JavaScript (ECMAScript 2021),
* możliwa transpilacja do starszej wersji np. ECMAScript 5:
  * według wyboru dewelopera: `tsc --target`,
  * problem wersji standardu i wersji przeglądarek.

---

### TypeScript i JavaScript

**TypeScript**:

* umożliwia korzystanie z istniejących bibliotek języka JavaScript,
* poprawny kod JavaScript jest też poprawnym kodem języka TypeScript
* declaration Files - zawierają deklaracje typów dla bibliotek zaimplementowanych w JavaScript:
  * np. <http://definitelytyped.org/>;
* Wygenerowane pliki .js są (w miarę) czytelne dla dewelopera.

---

### Podstawowe użycie

Instalacja narzędzia:

```bash
npm install -g typescript
```

Transpilacja pliku:

```bash
tsc file.ts
```

Obserwowanie zmian:

```bash
tsc --watch file.ts
```

Popularne IDE oferują integrację z narzędziami.

---

### Typy podstawowe

```typescript
let done: boolean = false;

let age: number = 42;

let color: number = 0xf00dcc;

let username: string = "ookami";

let list1: number[] = [42, 36, 28];

let list2: Array<number> = [27, 45, 19];
```

```javascript
let done = false;

let age = 42;

let color = 0xf00dcc;

let username = "ookami";

let list1 = [42, 36, 28];

let list2 = [27, 45, 19];
```

---

### Typy wyliczeniowe i krotki

```typescript
let x: [string, number] = ["ookami", 24];

enum Color {Red, Green, Blue}

let c: Color = Color.Green;
```

```javascript
let x = ["ookami", 24];

var Color;
(function (Color) {
  Color[Color["Red"] = 0] = "Red";
  Color[Color["Green"] = 1] = "Green";
  Color[Color["Blue"] = 2] = "Blue";
})(Color || (Color = {}));

let c = Color.Green;
```

---

### Dowolny typ

```typescript
let variable: any = 42;
variable = "ookami";
variable = false;
```

```javascript
let variable = 42;
variable = "ookami";
variable = false;
```

---

### Parametry funkcji

```typescript
function greeter(username: string): string {
    return "Hello, " + username;
}
```

```javascript
function greeter(username) {
  return "Hello, " + username;
}
```

---

### Parametry funkcji - opcjonalne

```typescript
function greeter(username?: string): string {
    if (username) {
        return "Hello, " + username;
    } else {
        return "Hello!"
    }
}
```

```javascript
function greeter(username) {
    if (username) {
        return "Hello, " + username;
    } else {
        return "Hello!";
    }
}
```

---

### Parametry funkcji - domyślne

```typescript
function greeter(user: string = "world"): string {
    return "Hello, " + user;
}
```

```javascript
function greeter(user = "world") {
    return "Hello, " + user;
}
```

---

### Parametry funkcji - zmienna liczba

```typescript
function greeter(firstName: string, ...otherNames: string[]): string {
    return "Hello, " + username + " " + otherNames.join(", ");
}

greeter("world", "ookami", "kitsune");
```

```javascript
function greeter(firstName, ...otherNames) {
  return "Hello, " + username + " " + otherNames.join(", ");
}

greeter("world", "ookami", "kitsune");
```

---

### Klasy - definicja

```typescript
class Greeter {

    greeting: string;

    constructor(message: string) {
        this.greeting = message;
    }

    greet(): string {
        return "Hello, " + this.greeting;
    }

}
```

```javascript
class Greeter {
    
  constructor(message) {
    this.greeting = message;
  }
  
  greet() {
    return "Hello, " + this.greeting;
  }
  
}
```

---

### Klasy - tworzenie obiektu

```typescript
let greeter: Greeter = new Greeter("world");
let greetings = greeter.greet();
```

```javascript
let greeter = new Greeter("world");
let greetings = greeter.greet();
```

---

### Klasy - klasa bazowa

```typescript
class Animal {

    public name: string;

    public constructor(name: string) {
        this.name = name;
    }

    public move(distanceInMeters: number) {
        console.log(`${this.name} moved ${distanceInMeters}m.`);
    }

}
```

```javascript
class Animal {
    
  constructor(name) {
    this.name = name;
  }
  
  move(distanceInMeters) {
    console.log(`${this.name} moved ${distanceInMeters}m.`);
  }
  
}
```

---

### Klasy - dziedziczenie

```typescript
class Snake extends Animal {

    constructor(name: string) {
        super(name);
    }

    move(distanceInMeters = 5) {
        console.log("Slithering...");
        super.move(distanceInMeters);
    }

}
```

```javascript
class Snake extends Animal {
    
  constructor(name) {
    super(name);
  }
  
  move(distanceInMeters = 5) {
    console.log("Slithering...");
    super.move(distanceInMeters);
  }
  
}
```

---

### Klasy - dziedziczenie cd

```typescript
class Horse extends Animal {

    constructor(name: string) {
        super(name);
    }

    move(distanceInMeters = 45) {
        console.log("Galloping...");
        super.move(distanceInMeters);
    }

}
```

```javascript
class Horse extends Animal {
    
  constructor(name) {
    super(name);
  }
  
  move(distanceInMeters = 45) {
    console.log("Galloping...");
    super.move(distanceInMeters);
  }
  
}
```

---

### Poziomy dostępu

**Poziomy dostępu**:

* public - domyślny, pola i metody widoczne dla innych klas,
* protected - widoczne tylko w obrębie hierarchii dziedziczenia,
* private - dostępne tylko w obrębie danej klasy.

---

### Poziomy dostępu - przykład

```typescript
class Animal {

    private _name: string;

    get name(): string {
        return this._name;
    }

    set name(newName: string) {
        this._name = newName;
    }

}
```

```javascript
class Animal {
    
  get name() {
    return this._name;
  }
  
  set name(newName) {
    this._name = newName;
  }
  
}
```

---

### Klasy abstrakcyjne

```typescript
abstract class Animal {

    abstract makeSound(): void;

    move(): void {
        console.log("move");
    }

}
```

```javascript
class Animal {
    
  move() {
    console.log("move");
  }
  
}
```

---

### Interfejsy

```typescript
interface Moveable {

    move(distanceInMeters: number): void;

}
```

```typescript

```

---

### Interfejsy - implementacja

```typescript
class Snake implements Moveable {

    move(distanceInMeters = 5) {
        console.log("Slithering...");
    }

}
```

```javascript
class Snake {
    
  move(distanceInMeters = 5) {
    console.log("Slithering...");
  }
  
}
```

---

### Interfejsy - własności

```typescript
interface Person {

    firstName: string;

    lastName: string;

    get email(): string;

}
```

```javascript

```

---

### Interfejsy - przykład

```typescript
class Student implements Person {

    firstName: string;

    lastName: string;

    get email(): string {
        return this.firstName
            + "." + this.lastName
            + "@student.pg.edu.pl";
    }

}
```

```javascript
class Student {
    
  get email() {
    return this.firstName
            + "." + this.lastName
            + "@student.pg.edu.pl";
  }
  
}
```

---

### Interfejsy - polimorfizm

```typescript
let person: Person = new Student();
```

```javascript
let person = new Student();
```

---

### Pomocne źródła

Dokumentacja:

* Microsoft, *TypeScript Documentation*, <https://www.typescriptlang.org/docs/>.

Uzupełnienie informacji:

* w3schools, *TypeSript Tutorial*, <https://www.w3schools.com/typescript/>.
