# Narzędzia i Aplikacje Jakarta EE

## Angular

Michał Wójcik

---

### Aplikacje internetowe

**Przed erą mobilną**:

* cała mechanika witryny po stronie serwera,
* przeglądarka wyświetla gotowe dokumenty HTML otrzymane od serwera,
* ograniczona dynamika w przeglądarce.

**Początek ery mobilnej**:

* serwer generuje dokumenty HTML dla przeglądarki, a dla aplikacji mobilnej udostępnia API (np. dane w formacie JSON),
* aplikacja serwerowa udostępnia swoją funkcjonalność na dwa różne sposoby - kłopotliwe, problemy z bezpieczeństwem.

**W międzyczasie**:

* wzrost wydajności przeglądarek,
* możliwość obsłużenia skomplikowanych interakcji.

---

### Aplikacje internetowe - API

Skoro **aplikacja serwerowa** i tak musi udostępniać API dla klienta mobilnego, to czy klient w przeglądarce nie mógłby
używać **tego samego API**?

**Te same punkty wejścia** dla wszystkich aplikacji klienckich:

* jednolite mechanizmy logowania i kontroli dostępu,
* aplikacja serwerowa skupia się na logice biznesowej,
* za interakcje z użytkownikiem odpowiedzialne są aplikacje klienckie,
* przejrzysty podział odpowiedzialności: front-end lub back-end deweloper.

---

### Budowanie złożonych aplikacji webowych

Istnieje duża liczba **rozwiązań** (zarówno frameworków, jak i bibliotek):

* Angular,
* React,
* Vue.js,
* Ember.js,
* Meteor,
* Mithril,
* Node.js,
* Polymer,
* Aurelia,
* Backbone.js.

---

### Angular jako framework

Angular jest frameworkiem, który ułatwiającym **budowanie złożonych interakcji** w przeglądarce:

* aplikacje mogą (nie muszą) spełniać model MVC (Model View Controller):
  * wzorzec architektoniczny modelujący przepływ sterowania w interfejsie użytkownika;
* definiuje, w jaki sposób zorganizować kod projektu:
  * jak podzielić logikę interakcji,
  * jak powiązać logikę z widokami;
* dostarczają funkcje narzędziowe do wykonywania typowych operacji:
  * np. wywoływanie usług po stronie serwera.

---

### Framework Angular

Wersje frameworku **Angular**:

* Angular 1.x - 2009,
* Angular 2.x - 2016,
  * framework został przepisany od zera,
  * język TypeScript,
  * wersja 2.x znacznie bardziej przypomina framework React niż Angular 1.x:
  * migracja 1.x → 2.x wymaga wielu modyfikacji;
* Angular 4, 5, 6, 7, 8, 9, ..., 14, 15, 16:
  * wersjonowanie semantyczne
  * wersja 3.x pominięta ze względu na rozbieżność wersji poszczególnych komponentów (router package),
  * Google planuje nowe wydanie średnio co pół roku,
  * 18 miesięcy wsparcia (6+12).

---

### Wersjonowanie semantyczne

Numery wersji w postaci: **major.minor.patch**.

Przykład: **2.7.3**.

**Elementy numeru wersji**:

* patch - poprawki błędów kompatybilne z poprzednią wersją, np. 2.7.4,
* minor - nowe funkcjonalności kompatybilne z poprzednimi wersjami, np. 2.8.0,
* major - zmiany niekompatybilne wstecznie, np. 3.0.

---

### Wersje frameworka a aplikacje

Jeśli chcemy mieć dostęp do **nowych funkcji oraz poprawek bezpieczeństwa**, musimy **migrować**
naszą aplikację na nowsze wersje frameworka:

* patch, minor - zmiany kompatybilne wsteczne:
  * kod nasz naszej aplikacji nie wymaga modyfikacji do poprawnego działania na nowej wersji frameworka;
* major - wymaga zmian w naszym kodzie, aby dokonać migracji:
  * wymaga roboczogodzin,
  * te zmiany nie muszą być bolesne dla deweloperów,
  * autorzy Angulara nie planują przepisywać frameworka do zera kolejny raz,
  * każde kolejne wydanie wprowadza nowe zmiany, jeśli zrezygnujemy z migracji, później będzie już tylko trudniej.

---

### Angular - podstawowe użycie

**Instalacja** narzędzia:

```bash
npm install -g @angular/cli
```

Stworzenie nowego projektu:

```bash
ng new project-name --routing --style=css

```

Uruchomienie aplikacji:

```bash
ng serve --open
```

---

### Struktura projektu

**Źródła**:

```tex
src\                (source files for the root-level application project)
    app\            (component files which application logic and data)
    assets\         (imaged and other assets)
    favicon.ico     (icon for bookmark)
    index.html      (main page)
    main.ts         (application entry point)
    styles.css      (global styles)
```

**Inne pliki**:

```tex
.editorconfig       (config for editors)
.gitignore          (git ignore configuration)
angular.json        (CLI config)
package.json        (list of npm packages dependencies)
package-lock.json   (list of resolved npm packages versions)
README.md           (just a readme)
tsconfig.app.json   (project specific TypeScript config)
tsconfig.json       (default TypeScript config)
tsconfig.spec.json  (TypeScript config for tests)
```

---

### Architektura

**Architektura**:

* aplikacja wykorzystuje język HTML i style CSS,
* poza standardowymi znacznikami HTML występują również dodatkowe specjalne znaczniki,
* dodatkowe znaczniki umożliwiają wykorzystanie komponentów definiowanych w projekcie:
  * komponent opisuje fragment widoku oraz jego zachowania,
  * komponent może być zbudowany z innych komponentów (hierarchiczna struktura widoków);
* logika aplikacji implementowana w serwisach,
* komponenty i serwisy grupowane w modułach.

---

### Architektura - elementy

**Elementy**:

* moduły (ang. modules),
* komponenty (ang. components),
* szablony (ang. templates),
* metadane (ang. metadata),
* wiązanie danych (ang. data binding),
* dyrektywy (ang. directives),
* serwisy (ang. services),
* wstrzykiwanie zależności (ang. dependency injection),
* przekształcenia (ang. pipes).

---

### Moduły

**Moduły**:

* aplikacje w Angularze dzieli się na moduły odpowiadające poszczególnym funkcjonalnościom
* każda aplikacja posiada główny moduł o nazwie `AppModule`,
* małe aplikacje mogą posiadać tylko jeden moduł, duże aplikacje - setki modułów;
* moduły definiowane jako klasy z dekoratorem `@NgModule`:
  * dekoratory umożliwiają dołączenie metadanych do klasy.

---

### NgModule - metadane

Najważniejsze **metadane** dekoratora `NgModule`:

* `declarations` - lista komponentów wykorzystywanych do budowania widoków aplikacji,
* `exports` - lista komponentów, które powinny być dostępne do wykorzystania przez inne moduły,
* `imports` - lista modułów, których wyeksportowane klasy są wykorzystywane w module bieżącym,
* `providers` - lista dostawców umożliwiających budowanie instancji serwisów do wykorzystania w całej aplikacji (we
  wszystkich modułach),
* `bootstrap` - komponent reprezentujący główny widok aplikacji (do niego ładowane są wszystkie inne widoki), używany
  tylko dla `AppModule`.

---

### NgModule - przykład

```typescript
import { BrowserModule } from '@angular/platform-browser';
import { NgModule } from '@angular/core';

import { AppComponent } from './app.component';

@NgModule({
    declarations: [
        AppComponent
    ],
    imports: [
        BrowserModule
    ],
    providers: [],
    bootstrap: [AppComponent]
})
export class AppModule {
}
```

---

### Moduły Angular vs JavaScript

**Moduły**:

* w języku JavaScript każdy plik jest modułem,
* moduły w języku JavaScript i moduły frameworka Angular to dwie osobne koncepcje,
* JavaScript: moduł = plik,
* Angular: moduł = zbiór komponentów i serwisów,
* w typowym projekcie używamy obu mechanizmów modułów równocześnie,
* wspólne nazewnictwo (moduł/import/export) dla odrębnych mechanizmów może wywoływać dezorientację.

---

### Moduły w języku JavaScript

Słowo kluczowe `export` w plikach (modułach) `.js` określa obiekty, które dany moduł udostępnia innym modułom (
plikom `.js`):

```javascript
export class AppModule {
}
```

```javascript
export class AppComponent {
}
```

Inne pliki (moduły) `.js` mogą korzystać z tych obiektów korzystając z dyrektywy `import`:

```javascript
import {AppComponent} from "./app.component";
```

---

### Uruchomienie aplikacji

**Uruchomienie aplikacji**:

* należy zaimportować klasę `AppModule` z pliku (modułu) JavaScript,
* wywołanie metody `bootstrapModule` inicjuje moduł Angulara:

```typescript
import { platformBrowserDynamic } from '@angular/platform-browser-dynamic';
import { AppModule } from './app/app.module';

platformBrowserDynamic().bootstrapModule(AppModule);
```

---

### Komponenty

**Komponenty**:

* reprezentują fragmenty widoków składające się na interfejs aplikacji,
* definiowane jako klasy z dekoratorem `@Component`,
* są reprezentowane przez dodatkowe znaczniki umieszczane w kodzie HTML, np.:

```html
<body>
    <app-root></app-root>
</body>
```

---

### Komponenty - elementy

**Komponent** określa:

* szablon używany do zbudowania fragmentu widoku (tagi HTML obejmujące też tagi innych komponentów),
* dane do prezentacji (model),
* zachowania (funkcje obsługi zdarzeń).

`@Component()` – najważniejsze metadane:

* `selector` – selektor CSS określający które znaczniki na stronie mają zostać wypełnione zawartością komponentu,
* `templateUrl` – ścieżka do pliku .html z szablonem,
* `providers` – lista dostawców serwisów, z których korzysta komponent.

`@Input()` - pozwala na przekazanie atrybutów z rodzica do komponentu.

`@Output()` - pozwala na przekazanie zdarzeń z komponentu do rodzica.

---

### Komponent - przykład

```typescript
import { Component, Input, OnInit } from '@angular/core';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    @Input()
    name: string;

    value: string;

    constructor() {
    }

    ngOnInit() {
        this.value = 'Hello ' + this.name + '!';
    }

}
```

```angular2html
<app-hello [name]="'world'"></app-hello>
```

---

### Komponent - deklaracja - przykład

```typescript
import { BrowserModule } from '@angular/platform-browser';
import { NgModule } from '@angular/core';

import { AppComponent } from './app.component';
import { HelloComponent } from './hello/hello.component';

@NgModule({
    declarations: [
        AppComponent,
        HelloComponent // !
    ],
    imports: [
        BrowserModule
    ],
    providers: [],
    bootstrap: [AppComponent]
})
export class AppModule {
}
```

---

### Szablony

**Szablony**:

* wykorzystywane przez komponenty do wygenerowania treści w oknie przeglądarki,
* składnia opiera się na składni języka HTML,
* zawierają dodatkowe elementy:
  * znaczniki umożliwiające dołączanie kolejnych komponentów do widoku,
  * dyrektywy umożliwiające sterowanie procesem generowania wynikowego kodu HTML.

---

### Szablon - przykład

```angular2html
<div *ngIf="value">
  {{value}}
</div>
```

---

### Wiązanie danych

**Wiązanie danych**:

* bez wykorzystania wiązania danych deweloper musi ręcznie umieszczać w drzewie DOM dane uzyskane
  w kodzie JavaScript (np. wyniki żądania AJAXowego),
* oraz ręcznie pobierać wartości z drzewa DOM w reakcji na zdarzenia emitowane przez użytkownika(np. `keyup`, `click` itd.),
* przykładowo: walidacja formularza
  * po kliknięciu przycisku przez użytkownika pobranie danych z pól, a następnie dodanie komunikatów błędów na stronie;
* wiązanie danych automatyzuje proces wymiany informacji pomiędzy komponentem i drzewem DOM.

---

### 4 formy wiązania danych

Dostępne są 4 **formy wiązania danych**:

* interpolacja wartości: `{{...}}`,
* wiązanie właściwości elementów DOM: `[property]`.
* wiązanie obsługi zdarzenia: `(event)`,
* dwukierunkowe wiązanie danych: `[(...)]`.

---

### Interpolacja

**Interpolacja**:

* umożliwia wyznaczenie wartości wykorzystywanej
  w widoku na podstawie wyrażenia, np.:

```angular2html
<h3>{{imgTitle}}</h3>
<img src="{{imgUrl}}">
```

* wyrażenie najczęściej odnosi się do pól/własności klasy komponentu,
* wyrażenie w nawiasach `{{...}}` jest konwertowane na ciąg znaków przed umieszczeniem w widoku.

---

### Wyrażenia

**Wyrażenia** (template expressions):

* mogą realizować dodatkowe operacje,
* nie powinny powodować efektów ubocznych,
* powinny być szybkie do wykonania,
* powinny być możliwie krótkie i proste,
* złożoną logikę należy umieścić w metodzie komponentu i wywoływać gotową metodę w wyrażeniu,
* wyrażenie powinno być idempotentne.

```angular2html
<p>Przekroczono próg o {{score - threshold}} punktów.</p>
```

---

### Wiązanie własności

**Wiązanie własności**:

* wartości wyrażeń można również wiązać z właściwościami elementów drzewa DOM oraz właściwościami
  komponentów,
* takie wiązanie jest jednokierunkowe:
* zmiana wartości wyrażenia powoduje zmianę wartości właściwości, ale nie odwrotnie;
* wiązania odnoszą się do właściwości elementów drzewa DOM, a nie do atrybutów tagów HTML.

```angular2html
<button [disabled]="unchanged">Cancel</button>
<img [src]="imgUrl"/>
<app-book-detail [book]="selectedBook"/>
```

---

### Właściwości DOM a atrybuty HTML

**DOM** vs **HTML**:

* niektóre atrybuty elementów HTML mają bezpośrednie odzwierciedlenie we właściwościach węzłów drzew DOM, np. `id`, `src`,
* dla niektórych atrybutów nie istnieją odpowiadające im właściwości, np. `colspan`, `aria`,
* niektóre właściwości nie posiadają odpowiadających im atrybutów, np. `textContent`,
* niektóre atrybuty mają odzwierciedlenie we właściwościach, ale jest ono nieintuicyjne np. `disabled`:
  * w HTML – sama obecność atrybutu powoduje nieaktywność elementu (bez względu na wartość)
  * w drzewie DOM - `disabled=true` powoduje nieaktywność.

---

### Wiązanie dla atrybutów

**Wiązania** dla **atrybutów**:

* wiązania odnoszą się do właściwości elementów drzewa DOM a nie do atrybutów tagów HTMLowych,
* jeśli atrybut nie posiada odpowiednika wśród właściwości DOM jedyny sposób na jego zmianę, to
  bezpośrednia zmiana atrybutu – prefiks: `attr`.

```angular2html
<td [attr.colspan]="colsCount">One-Two</td>
```

---

### Wiązanie właściwości CSS

**Wiązanie** właściwości **CSS**:

* obiektem wiązania mogą być klasy CSS,
* albo poszczególne właściwości CSS.

```angular2html
<div [class.special]="special">...</div>
```

```angular2html
<button [style.color]="special ? 'red': 'green'">
<button [style.background-color]="canSave ? 'cyan':'grey'">
```

---

### Interpolacja czy wiązanie własności?

**{{}}** vs **[property]='...'**:

* interpolacja `{{...}}` zawsze konwertuje wynik na postać tekstową,
* wiązanie właściwości `[property]` zachowuje typ wiązanego wyrażenia.

---

### Wiązanie zdarzeń

**Wiązanie zdarzeń** - umożliwia wywoływanie funkcji obsługi zdarzeń zdefiniowanych w klasie
komponentu w reakcji na akcje użytkownika.

```angular2html
<button (click)="onSave()">Save</button>
```

---

### Wiązanie zdarzeń - przykład

```typescript
import { Component, Input, OnInit } from '@angular/core';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    constructor() {
    }

    ngOnInit() {
    }

    onSave(): void {

    }

}
```

---

### Wiązanie dwukierunkowe

**Wiązanie dwukierunkowe**:

* zmiana wartości związanego pola zmienia wartość właściwości,
* zmiana wartości właściwości zmienia wartość pola,
* szczególnie przydatne przy pracy z formularzami:
* budowane ręcznie lub poprzez `ngModel`.

```angular2html
<input [value]="name" (input)="name=$event.target.value" >
```

```angular2html
<input [(ngModel)]="name">
```

---

### Wysłanie zdarzenia - przykład

```typescript
import { Component, Input, OnInit } from '@angular/core';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    @Output()
    helloEvent: EventEmitter<string> = new EventEmitter<string>();

    constructor() {
    }

    ngOnInit() {
    }

    emit(): void {
        this.helloEvent.emit('hello');
    }

}
```

---

### Obsługa zdarzenia - przykład

```angular2html
<app-hello (helloEvent)="onHelloAction($event)"></app-hello>
```

---

### Dyrektywa ngModel

**ngModel**:

* reprezentuje wartość pola w formularzu:
  * umożliwia np. przypisanie wartości początkowej;
* równocześnie pozwala nasłuchiwać na zmiany:
  * zmiana wartości wykonana przez użytkownika jest automatycznie odzwierciedlona w polu klasy komponentu;
* obsługuje wszystkie podstawowe elementy formularzy HTML,
* ukrywa różnice pomiędzy nimi, np.:
  * `input` → `value`,
  * `select`/`option` → `selected`.

---

### Dyrektywy

**Dyrektywy**:

* dokumenty HTML posiadają statyczną strukturę,
* szablony widoków Angulara są dynamiczne,
* wynikowy HTML jest efektem przetworzenia szablonu zgodnie z umieszczonymi w nim dyrektywami,
* dwie grupy dyrektyw:
  * strukturalne - modyfikują strukturę dokumentu poprzez dodawanie, usuwanie lub zamianę elementów,
  * w postaci atrybutów – modyfikują wygląd lub zachowanie istniejących już elementów.

Przykładowe dyrektywy strukturalne:

* `\*ngFor` - dodawanie elementów w pętli,
* `\*ngIf` - warunkowe wyświetlanie elementów.

Przykładowa dyrektywa w postaci atrybuty:

* `ngModel` - wiązanie danych formularzy.

---

### Serwisy

**Serwisy**:

* logika aplikacji nie powinna być implementowana w klasach komponentów,
* komponent działa w kontekście konkretnego szablonu widoku - trudno ponownie wykorzystać logikę zaszytą w klasie
  komponentu w innych miejscach aplikacji;
* kontroler powinien definiować pola i metody na potrzeby wiązania danych, a logikę delegować do serwisów,
* serwisy implementują logikę aplikacji w sposób niezależny od interfejsu,
* łatwe wykorzystanie w wielu różnych kontekstach,
* serwisy są dostarczane do komponentów na drodze wstrzykiwania zależności.

---

### Wstrzykiwanie

**Wstrzykiwanie zależności** - Klasa nie odpowiada za pozyskanie swoich zależności, zamiast tego zależności są do niej
dostarczane z zewnątrz.

**Wstrzykiwanie w Angularze**:

* wstrzykiwanie zależności jest nieodzownym mechanizmem frameworka Angular:
  * wykorzystywane wewnętrznie przez sam framework,
  * oraz w komponentach opracowywanych przez autora aplikacji opartej na Angularze;
* za dostarczenie zależności odpowiada komponent `injector`:
  * `injector` utrzymuje instancje serwisów, które mogą być wstrzykiwane,
  * instancje pozyskiwane są przy użyciu dostawców (ang. providers),
  * należy w projekcie zdefiniować dostawców dla wymaganych serwisów;
* wstrzykiwanie przez parametry konstruktora.

---

### Serwis - przykład

```typescript
import { Injectable } from '@angular/core';

@Injectable()
export class HelloService {

    constructor() {
    }
}
```

```typescript
import { Component, Inject, Input, OnInit } from '@angular/core';
import { HelloService } from '../hello.service';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    constructor(private service: HelloService) {
    }

    ngOnInit() {
    }

}
```

---

### Serwis - konfiguracja - przykład

```typescript
import { BrowserModule } from '@angular/platform-browser';
import { NgModule } from '@angular/core';

import { AppComponent } from './app.component';
import { HelloComponent } from './hello/hello.component';
import { HelloService } from './hello.service';

@NgModule({
    declarations: [
        AppComponent,
        HelloComponent
    ],
    imports: [
        BrowserModule
    ],
    providers: [HelloService], // !
    bootstrap: [AppComponent]
})
export class AppModule {
}
```

---

### Przekształcenia

**Przekształcenia** pozwalają na modyfikowanie danych na potrzeby wyświetlania.

Standardowe przekształcenia:

* `DatePipe`,
* `UpperCasePipe`,
* `LowerCasePipe`,
* `CurrencyPipe`,
* `DecimalPipe`,
* `PercentPipe`.

Istnieje możliwość definiowania własnych przekształceń.

---

### Przekształcenia - przykład

```angular2html
<p>Book should be returned before {{deadline | date:'yyyy-MM-dd'}}.</p>
<p>Current penalty: {{penalty | currency:'EUR'}}</p>
```

---

### Własne przekształcenie

```typescript
import { Pipe, PipeTransform } from '@angular/core';

@Pipe({name: 'exponentialStrength'})
export class ExponentialStrengthPipe implements PipeTransform {

    transform(value: number, exponent = 1): number {
        return Math.pow(value, exponent);
    }

}
```

---

### Routing

**Routing**:

* typowe aplikacje internetowe składają się z wielu widoków, pomiędzy którymi nawiguje użytkownik,
* `RouterModule` umożliwia definiowanie adresów, które spowodują wyświetlenie wybranych komponentów(widoków) aplikacji,
* Znacznik `<base href="/">` w sekcji `<head>` pliku `index.html` określa bazową ścieżkę dla adresów w ramach aplikacji.

---

### Routing - przykład

```typescript
import { NgModule } from '@angular/core';
import { RouterModule, Routes } from '@angular/router';
import { HelloComponent } from './hello/hello.component';

const routes: Routes = [
    {path: 'hello', component: HelloComponent},
    {path: 'hello/:name', component: HelloComponent}
];

@NgModule({
    imports: [RouterModule.forRoot(routes)],
    exports: [RouterModule]
})
export class AppRoutingModule {
}
```

```angular2html
<body>
    <section>
        <router-outlet></router-outlet>
    </section>
</body>
```

Komponent zdefiniowany w routingu zostanie wyświetlony poniżej elementu: `<router-outlet></router-outlet>`.

---

### Routing - konfiguracja - przykład

```typescript
import { BrowserModule } from '@angular/platform-browser';
import { NgModule } from '@angular/core';

import { AppComponent } from './app.component';
import { HelloComponent } from './hello/hello.component';
import { HelloService } from './hello.service';
import { AppRoutingModule } from './app-routing.module'; // !

@NgModule({
    declarations: [
        AppComponent,
        HelloComponent
    ],
    imports: [
        BrowserModule,
        AppRoutingModule // !
    ],
    providers: [HelloService],
    bootstrap: [AppComponent]
})
export class AppModule {
}
```

---

### Nawigacja

```angular2html
<nav class="navbar">
    <ul>
        <li>
            <a routerLink="/hello">Hello</a>
        </li>
    </ul>
</nav>
<section>
    <router-outlet></router-outlet>
</section>
```

---

### Programistyczna nawigacja

```typescript
import { Component, Input, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    value: string;

    constructor(private route: ActivatedRoute, private router: Router) { // !
    }

    back() {
        this.router.navigateByUrl('/hello'); // !
    }
}
```

---

### Nawigacja z parametrem

```angular2html
<nav class="navbar">
    <ul>
        <li>
            <a routerLink="/hello">Hello</a>
        </li>
        <li>
            <a [routerLink]="['/hello', 'ookami']">Hello ookami</a>
        </li>
    </ul>
</nav>
<section>
    <router-outlet></router-outlet>
</section>
```

---

### Programistyczna nawigacja z parametrem

```typescript
import { Component, Input, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    value: string;

    constructor(private route: ActivatedRoute, private router: Router) { // !
    }

    back() {
        this.router.navigateByUrl('/hello');
    }

    refresh() {
        this.router.navigate(['hello', 'ookami']); // !
    }
}
```

---

### Obsługa parametrów nawigacji

```typescript
import { Component, Input, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';

@Component({
    selector: 'app-hello',
    templateUrl: './hello.component.html',
    styleUrls: ['./hello.component.css']
})
export class HelloComponent implements OnInit {

    value: string;

    constructor(private route: ActivatedRoute, private router: Router) { // !
    }

    ngOnInit() {
        const id = this.route.snapshot.paramMap.get('name'); // !
        if (id == null) {
            this.value = 'Hello!';
        } else {
            this.value = 'Hello ' + id + '!';
        }
    }

}
```

---

### Konsumowanie usług sieciowych

**Wykorzystanie usług sieciowych**:

* dane prezentowane w aplikacji front-endowej typowo pobierane są z serwera (z back-endu),
* informacje wprowadzane przez użytkownika są zapisywane na serwerze:
  * dane z formularzy na stronie, zawartość koszyka/zamówienia itd.;
* aplikacja back-endowa udostępnia swoje funkcje w postaci usług sieciowych (ang. web services):
  * np. w architekturze REST;
* serwis `HttpClient` umożliwia wysyłanie żądań HTTP do back-endu.

---

### Wywołanie usługi sieciowej - konfiguracja

```typescript
import { BrowserModule } from '@angular/platform-browser';
import { NgModule } from '@angular/core';

import { AppComponent } from './app.component';
import { HelloComponent } from './hello/hello.component';
import { HelloService } from './hello.service';
import { AppRoutingModule } from './app-routing.module';
import { HttpClientModule } from '@angular/common/http'; // !

@NgModule({
    declarations: [
        AppComponent,
        HelloComponent
    ],
    imports: [
        BrowserModule,
        AppRoutingModule,
        HttpClientModule // !
    ],
    providers: [HelloService],
    bootstrap: [AppComponent]
})
export class AppModule {
}
```

---

### Wywołanie usługi sieciowej - proxy

Aby uniknąć ustawiania na stałe adresu usługi, warto skorzystać z pliku `proxy.conf.json`:

```json
{
  "/api": {
    "target": "http://localhost:8080/jax-rs-1.0-SNAPSHOT/",
    "secure": false
  }
}
```

Konfiguracja serwera deweloperskiego.

---

### Wywołanie usługi sieciowej - przykład

```typescript
import { Injectable } from '@angular/core';
import { HttpClient } from '@angular/common/http';
import { Observable } from 'rxjs';
import { Book } from '../model/book';
import { Author } from '../model/author';

@Injectable()
export class BookService {

    constructor(private http: HttpClient) {
    }

    findAllBooks(): Observable<Book[]> {
        return this.http.get<Book[]>('api/books');
    }

}
```

---

### Polecenia ng

Zamiast tworzyć poszczególne elementy ręcznie, można skorzystać z odpowiednich poleceń:

Stworzenie zwykłej klasy (modelu):

```bash
ng generate class model/book
```

Stworzenie typu enum:

```bash
ng generate enum model/cover
```

Stworzenie komponentu:

```bash
ng generate component component/BookList
```

Stworzenie serwisu:

```bash
ng generate service service/book
```

---

### Polecenia ng cd

Stworzenie modułu routingu:

```bash
ng generate module app-routing --flat --module=app
```

Uruchomienie serwera:

```bash
ng serve
```

Uruchomienie serwera z wykorzystaniem pliku proxy:

```bash
ng serve --proxy-config proxy.conf.json
```

Ściągnięcie zależności w projekcie bez `node_modules`:

```bash
npm install
```

---

### Polecenia ng - konfiguracja proxy

Konfigurację proxy warto umieścić w pliku `package.json`, z którego korzysta m.in. Intelij:

```json
{
  "name": "angular",
  "version": "0.0.0",
  "scripts": {
    "ng": "ng",
    "start": "ng serve --proxy-config proxy.conf.json",
    "build": "ng build",
    "test": "ng test",
    "lint": "ng lint",
    "e2e": "ng e2e"
  }
}
```

---

### Pomocne źródła

Pomocne źródła:

* Google, *Angular Documentation*, <https://angular.io/docs>.
* Google, *Angular Tutorials*, <https://angular.io/tutorial>.
* Microsoft, *TypeScript Documentation*, <https://www.typescriptlang.org/docs/>.
* W3 Schools, *TypeScript Tutorial*, <https://www.w3schools.com/typescript/>.
