# Narzędzia i Aplikacje Jakarta EE

## Angular - Material Design

Michał Wójcik

---

### Material Design

**Material Design**:

* System projektowania stworzony przez Google.
* Został zaprojektowany w celu zapewnienia spójnego wyglądu i zachowania interfejsów użytkownika na różnych platformach
  i urządzeniach.
* Charakteryzuje się czystym i intuicyjnym designem, wykorzystującym cienie, światło oraz realistyczne efekty wizualne.
* Podstawowe zasady Material Design obejmują "Material", który symbolizuje powierzchnie i obiekty, oraz "Motion", co
  odnosi się do płynnych animacji i przejść.
* Elementy takie jak karty, przyciski, ikony, typografia i kolory są starannie zdefiniowane, aby zachować spójność i
  łatwość używania dla użytkowników.
* Projektanci wykorzystują Material Design, aby tworzyć aplikacje i strony internetowe, które są zarówno estetyczne, jak
  i funkcjonalne.
* Ma na celu ułatwienie interakcji użytkownika z interfejsem, zapewniając jednocześnie pozytywne doświadczenia wizualne.

---

### Implementacje

Implementacje:

* Material Design Lite - niezależna biblioteka JavaScript od Google, nierozwijana, <https://getmdl.io/>,
* Material Components for the web - następca Material Design Lite,
* Material Web - zbudowane na Material Components for the web, dostarcza niestandardowe elementy HTML,
* Materialize - prosta i lekka biblioteka JavaScript, nierozwijana, <https://materializecss.com/>,
* Materialize - fork oryginalnego Materialize, <https://materializeweb.com/>,
* Material UI - dopracowana implementacja dla technologii React,<https://mui.com/>,
* MUI CSS - lekki framework CSS, <https://www.muicss.com/>,
* Material Design for Boostrap - dopracowana biblioteka łączą material design i boostrap, <https://mdbootstrap.com/>:
  * plain JavaScript,
  * Angular,
  * React,
  * Vue;
* Angular Material - biblioteka specjalnie dla technologii Angular <https://material.angular.io/>.

---

### Instalacja

Instalacja i konfiguracja za pomocą schematu instalacji (install schematics).

```bash
ng add @angular/material
```

Spowoduje to nie tylko zainstalowanie zależności ale także odpowiednie skonfigurowanie projektu (np.: dodanie
odpowiednich styli).

---

### Moduły

Każdy element interfejsu ma odpowiadający mu moduł, przykładowo:

* `MatButtonModule` - przyciski,
* `MatTableModule` - tabele,
* `MatInputModule` - kontrolki formularza.

Włączenie animacji za pomocą `BrowserAnimationsModule`.

---

### Komponenty - atrybuty

Część komponentów jest zdefiniowana za pomocą atrybutów dodawanych do standardowych elementów.

```angular2html
<button mat-raised-button color="primary">Primary</button>
```

Atrybut `color` wskazuje na jeden z podstawowych kolorów zdefiniowanych w głównym szablonie: `basic` (
domyślny), `primary`, `accent` i `warn`.

---

### Komponenty - kontener

Niektóre komponenty wykorzystują `ng-container` wraz z dyrektywami.

```angular2html
<table mat-table [dataSource]="books">

    <ng-container matColumnDef="position">
        <th mat-header-cell *matHeaderCellDef> Title</th>
        <td mat-cell *matCellDef="let element"> {{element.title}} </td>
    </ng-container>

    <ng-container matColumnDef="name">
        <th mat-header-cell *matHeaderCellDef> Author</th>
        <td mat-cell *matCellDef="let element"> {{element.author}} </td>
    </ng-container>

    <tr mat-header-row *matHeaderRowDef="displayedColumns"></tr>
    <tr mat-row *matRowDef="let row; columns: displayedColumns;"></tr>
</table>
```

---

### Komponenty

Cześć komponentów jest realizowana w postaci nowych tagów.

```angular2html
<mat-checkbox [(ngModel)]="book.isHardcover">
    Hardcover
</mat-checkbox>
```

---

### Elementy formularza

Wykorzystanie `mar-form-field` pozwala na zbudowanie spójnego (style, animacje) pola formularza oraz dodanie do niego
dodatkowych elementów (etykieta, walidacja itd.).

```angular2html
<form>
    <mat-form-field>
        <mat-label>Title</mat-label>
        <input matInput [formControl]="titleFormControl" placeholder="Ruda Sfora">
        <mat-hint>Errors will appear here</mat-hint>
        <mat-error *ngIf="emailFormControl.hasError('required')">
            Title is <strong>required</strong>
        </mat-error>
    </mat-form-field>
</form>
```

---

### Pomocne źródła

Pomocne źródła:

* <https://material.angular.io/guides>
* <https://material.angular.io/components/categories>
